from typing import Any, Callable, Dict, List, Tuple

import wrapt
from django.apps.config import AppConfig
from django.contrib.admin.sites import AdminSite
from django.http import HttpRequest
from django.urls import URLPattern, path, reverse

from django_managerie.blocklist import COMMAND_BLOCKLIST
from django_managerie.commands import ManagementCommand, get_commands
from django_managerie.types import CommandMap


def user_is_superuser(request: HttpRequest) -> bool:
    """
    Return True if the user is an active superuser, False otherwise.
    """
    return bool(request.user.is_active and getattr(request.user, "is_superuser", None))


class Managerie:
    ignored_app_names = {
        'django.core',
        'django.contrib.staticfiles',
    }

    def __init__(self, admin_site: AdminSite) -> None:
        self.admin_site = admin_site

    def patch(self) -> None:
        if hasattr(self.admin_site, 'patched_by_managerie'):
            return
        old_get_app_list = self.admin_site.get_app_list
        self.admin_site.get_app_list = self.patched_get_app_list(old_get_app_list)  # type: ignore[assignment]
        old_get_urls = self.admin_site.get_urls
        self.admin_site.get_urls = self.patched_get_urls(old_get_urls)  # type: ignore[assignment]
        self.admin_site.patched_by_managerie = True  # type: ignore[attr-defined]

    def is_command_allowed(
        self,
        request: HttpRequest,
        command: ManagementCommand,
    ) -> bool:
        """
        Return True if the command is allowed to be run in the current request.

        The default implementation checks if the command is enabled and then
        whether the user is a superuser.

        This can be overridden to implement per-command permissions.
        """
        if not self.is_command_enabled(command):
            return False
        return user_is_superuser(request)

    def is_command_enabled(self, command: ManagementCommand) -> bool:
        """
        Return True if the command is enabled (not blocklisted or opt-outed).

        This is only called by `is_command_allowed` as a pre-check before
        checking for per-request permissions.
        """
        if command.full_name in COMMAND_BLOCKLIST:
            return False
        if getattr(command.get_command_class(), 'disable_managerie', False):
            return False
        return True

    def get_commands(
        self,
        request: HttpRequest,
    ) -> Dict[AppConfig, CommandMap]:
        command_map: Dict[AppConfig, CommandMap] = {}
        for app_config, commands in get_commands().items():
            command_map[app_config] = {
                command_name: command
                for (command_name, command) in commands.items()
                if self.is_command_allowed(command=command, request=request)
            }
        return command_map

    def get_commands_for_app_label(
        self,
        request: HttpRequest,
        app_label: str,
    ) -> CommandMap:
        for app_config, commands in self.get_commands(request).items():
            if app_config.label == app_label:
                return commands
        return {}

    @wrapt.decorator
    def patched_get_app_list(self, wrapped, instance: AdminSite, args, kwargs):
        request = args[0]
        app_list: List[Dict] = wrapped(*args, **kwargs)
        if request.user.is_superuser:
            self._augment_app_list(request, app_list)
        return app_list

    @wrapt.decorator
    def patched_get_urls(
        self,
        wrapped: Callable,
        instance: AdminSite,
        args: Tuple[()],
        kwargs: Dict[Any, Any],
    ) -> list:
        urls = wrapped(*args, **kwargs)
        return self._get_urls() + list(urls)

    def _augment_app_list(
        self,
        request: HttpRequest,
        app_list: List[Dict],
    ):
        all_commands: Dict[str, CommandMap] = {
            app_config.label: commands
            for (app_config, commands) in self.get_commands(request).items()
        }
        for app in app_list:
            if all_commands.get(app['app_label']):  # Has commands?
                app.setdefault('models', []).append(self._make_app_command(app))

    def _make_app_command(self, app):
        return {
            'perms': {'change': True},
            'admin_url': reverse(
                'admin:managerie_list',
                kwargs={'app_label': app['app_label']},
                current_app=self.admin_site.name,
            ),
            'name': 'Commands',
            'object_name': '_ManagerieCommands_',
        }

    def _get_urls(self) -> List[URLPattern]:
        from django_managerie.views import ManagerieCommandView, ManagerieListView

        return [
            path(
                'managerie/<app_label>/<command>/',
                ManagerieCommandView.as_view(managerie=self),
                name='managerie_command',
            ),
            path(
                'managerie/<app_label>/',
                ManagerieListView.as_view(managerie=self),
                name='managerie_list',
            ),
            path(
                'managerie/',
                ManagerieListView.as_view(managerie=self),
                name='managerie_list_all',
            ),
        ]

    @property
    def urls(self):
        return (self._get_urls(), 'admin', self.admin_site.name)



import io
import time
import traceback
from contextlib import redirect_stderr, redirect_stdout
from itertools import chain
from typing import Any, Dict, Iterable, Optional

from django.apps import apps
from django.apps.config import AppConfig
from django.http import Http404, HttpResponse
from django.views.generic import FormView, TemplateView

from django_managerie.commands import ManagementCommand
from django_managerie.forms import ArgumentParserForm
from django_managerie.managerie import Managerie


class MenagerieBaseMixin:
    managerie: Optional[Managerie] = None
    kwargs: Dict[str, Any]
    _app: Optional[AppConfig]

    def get_app(self) -> Optional[AppConfig]:
        if hasattr(self, '_app'):
            return self._app
        if 'app_label' in self.kwargs:
            self._app = apps.get_app_config(self.kwargs['app_label'])
            return self._app
        return None


class ManagerieListView(MenagerieBaseMixin, TemplateView):
    template_name = 'django_managerie/admin/list.html'

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context['app'] = app = self.get_app()
        context['title'] = f"{app.verbose_name if app else 'All Apps'} – Commands"
        managerie = self.managerie
        assert managerie
        commands: Iterable[ManagementCommand]
        if app:
            commands = managerie.get_commands_for_app_label(
                request=self.request,
                app_label=app.label,
            ).values()
        else:
            commands = chain(
                *(
                    app_commands.values()
                    for app_commands in managerie.get_commands(
                        request=self.request
                    ).values()
                )
            )
        context['commands'] = sorted(commands, key=lambda cmd: cmd.full_title)
        return context


class ManagerieCommandView(MenagerieBaseMixin, FormView):
    template_name = 'django_managerie/admin/command.html'

    @property
    def command_name(self) -> str:
        return self.kwargs['command']

    def get_command_object(self) -> ManagementCommand:
        app = self.get_app()
        managerie = self.managerie
        assert app and managerie
        try:
            return managerie.get_commands_for_app_label(
                request=self.request,
                app_label=app.label,
            )[self.command_name]
        except KeyError:
            raise Http404(
                f"Command {self.command_name} not found in {app.label} "
                f"(or you don't have permission to run it)"
            )

    def get_form(self, form_class=None) -> ArgumentParserForm:
        cmd = self.get_command_object().get_command_instance()
        parser = cmd.create_parser('django', self.command_name)
        return ArgumentParserForm(parser=parser, **self.get_form_kwargs())

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)
        command = self.get_command_object()
        context.update(
            app=self.get_app(),
            command=command,
            command_help=command.get_command_instance().help,
            title=command.full_title,
        )
        return context

    def form_valid(self, form: ArgumentParserForm) -> HttpResponse:
        # This mimics BaseCommand.run_from_argv():
        options = dict(form.cleaned_data)
        # "Move positional args out of options to mimic legacy optparse"
        args = options.pop('args', ())
        stdout = io.StringIO()
        stderr = io.StringIO()
        error = None
        error_tb = None
        t0 = time.time()
        co = self.get_command_object()
        with redirect_stdout(stdout), redirect_stderr(stderr):
            options.update(
                {
                    'traceback': True,
                    'no_color': True,
                    'force_color': False,
                    'stdout': stdout,
                    'stderr': stderr,
                }
            )
            cmd = co.get_command_instance()
            try:
                cmd._managerie_request = self.request
                cmd.execute(*args, **options)
            except SystemExit as se:  # We don't want any stray sys.exit()s to quit the app server
                stderr.write(f'<exit: {se}>')
            except Exception as exc:
                error = exc
                error_tb = traceback.format_exc()
        context = self.get_context_data(
            executed=True,
            form=form,
            error=error,
            error_tb=error_tb,
            stdout=stdout.getvalue(),
            stderr=stderr.getvalue(),
            duration=(time.time() - t0),
        )
        return self.render_to_response(context=context, status=(400 if error else 200))
